@@ -2,7 +2,7 @@ use std::task::{Context, Poll};
 
 use bytes::Buf;
 
-use futures_util::ready;
+#[cfg(feature = "tracing")]
 use tracing::trace;
 
 use crate::stream::{BufRecvStream, WriteBuf};
@@ -96,7 +96,11 @@ where
             return Poll::Ready(Ok(None));
         };
 
-        let end = ready!(self.try_recv(cx))?;
+        let end = match self.try_recv(cx) {
+            Poll::Ready(Ok(end)) => end,
+            Poll::Ready(Err(e)) => return Poll::Ready(Err(e)),
+            Poll::Pending => false,
+        };
         let data = self.stream.buf_mut().take_chunk(self.remaining_data);
 
         match (data, end) {
@@ -224,8 +228,10 @@ impl FrameDecoder {
             };
 
             match decoded {
-                Err(frame::FrameError::UnknownFrame(ty)) => {
-                    trace!("ignore unknown frame type {:#x}", ty);
+                Err(frame::FrameError::UnknownFrame(_ty)) => {
+                    #[cfg(feature = "tracing")]
+                    trace!("ignore unknown frame type {:#x}", _ty);
+
                     src.advance(pos);
                     self.expected = None;
                     continue;
