@@ -7,9 +7,11 @@ use assert_matches::assert_matches;
 use bytes::{Buf, Bytes, BytesMut};
 use futures_util::future;
 use http::{Request, Response, StatusCode};
+use tokio::sync::oneshot::{self};
 
+use crate::client::SendRequest;
+use crate::{client, server};
 use crate::{
-    client::{self, SendRequest},
     connection::ConnectionState,
     error::{Code, Error, Kind},
     proto::{
@@ -20,7 +22,6 @@ use crate::{
         varint::VarInt,
     },
     quic::{self, SendStream},
-    server,
 };
 
 use super::h3_quinn;
@@ -32,15 +33,16 @@ async fn connect() {
     let mut server = pair.server();
 
     let client_fut = async {
-        let _ = client::new(pair.client().await).await.expect("client init");
+        let (mut drive, _client) = client::new(pair.client().await).await.expect("client init");
+        future::poll_fn(|cx| drive.poll_close(cx)).await.unwrap();
     };
 
     let server_fut = async {
         let conn = server.next().await;
-        let _ = server::Connection::new(conn).await.unwrap();
+        let _server = server::Connection::new(conn).await.unwrap();
     };
 
-    tokio::join!(server_fut, client_fut);
+    tokio::select!(() = server_fut => (), () = client_fut => panic!("client resolved first"));
 }
 
 #[tokio::test]
@@ -48,14 +50,21 @@ async fn accept_request_end_on_client_close() {
     let mut pair = Pair::default();
     let mut server = pair.server();
 
+    let (tx, rx) = oneshot::channel::<()>();
+
     let client_fut = async {
-        let _ = client::new(pair.client().await).await.expect("client init");
+        let client = pair.client().await;
+        let client = client::new(client).await.expect("client init");
+        // wait for the server to accept the connection
+        rx.await.unwrap();
         // client is dropped, it will send H3_NO_ERROR
+        drop(client);
     };
 
     let server_fut = async {
         let conn = server.next().await;
         let mut incoming = server::Connection::new(conn).await.unwrap();
+        tx.send(()).unwrap();
         // Accept returns Ok(None)
         assert!(incoming.accept().await.unwrap().is_none());
     };
@@ -65,6 +74,7 @@ async fn accept_request_end_on_client_close() {
 
 #[tokio::test]
 async fn server_drop_close() {
+    init_tracing();
     let mut pair = Pair::default();
     let mut server = pair.server();
 
@@ -73,8 +83,8 @@ async fn server_drop_close() {
         let _ = server::Connection::new(conn).await.unwrap();
     };
 
-    let (mut conn, mut send) = client::new(pair.client().await).await.expect("client init");
     let client_fut = async {
+        let (mut conn, mut send) = client::new(pair.client().await).await.expect("client init");
         let request_fut = async move {
             let mut request_stream = send
                 .send_request(Request::get("http://no.way").body(()).unwrap())
@@ -93,8 +103,45 @@ async fn server_drop_close() {
     tokio::join!(server_fut, client_fut);
 }
 
+// In this test the client calls send_data() without doing a finish(),
+// i.e client keeps the body stream open. And client expects server to
+// read_data() and send a response
+#[tokio::test]
+async fn server_send_data_without_finish() {
+    let mut pair = Pair::default();
+    let mut server = pair.server();
+
+    let client_fut = async {
+        let (_driver, mut send_request) = client::new(pair.client().await).await.unwrap();
+
+        let mut req = send_request
+            .send_request(Request::get("http://no.way").body(()).unwrap())
+            .await
+            .unwrap();
+        let data = vec![0; 100];
+        req.send_data(bytes::Bytes::copy_from_slice(&data))
+            .await
+            .unwrap();
+        let _ = req.recv_response().await.unwrap();
+    };
+
+    let server_fut = async {
+        let conn = server.next().await;
+        let mut incoming = server::Connection::new(conn).await.unwrap();
+        let (_, mut stream) = incoming.accept().await.unwrap().unwrap();
+        let mut data = stream.recv_data().await.unwrap().unwrap();
+        let data = data.copy_to_bytes(data.remaining());
+        assert_eq!(data.len(), 100);
+        response(stream).await;
+        server.endpoint.wait_idle().await;
+    };
+
+    tokio::join!(server_fut, client_fut);
+}
+
 #[tokio::test]
 async fn client_close_only_on_last_sender_drop() {
+    init_tracing();
     let mut pair = Pair::default();
     let mut server = pair.server();
 
@@ -109,18 +156,24 @@ async fn client_close_only_on_last_sender_drop() {
     let client_fut = async {
         let (mut conn, mut send1) = client::new(pair.client().await).await.expect("client init");
         let mut send2 = send1.clone();
-        let _ = send1
+        let mut request_stream_1 = send1
             .send_request(Request::get("http://no.way").body(()).unwrap())
             .await
-            .unwrap()
-            .finish()
-            .await;
-        let _ = send2
+            .unwrap();
+
+        let _ = request_stream_1.recv_response().await;
+
+        let _ = request_stream_1.finish().await;
+
+        let mut request_stream_2 = send2
             .send_request(Request::get("http://no.way").body(()).unwrap())
             .await
-            .unwrap()
-            .finish()
-            .await;
+            .unwrap();
+
+        let _ = request_stream_2.recv_response().await;
+
+        let _ = request_stream_2.finish().await;
+
         drop(send1);
         drop(send2);
 
@@ -330,13 +383,27 @@ async fn control_close_send_error() {
         //# If either control
         //# stream is closed at any point, this MUST be treated as a connection
         //# error of type H3_CLOSED_CRITICAL_STREAM.
-        control_stream.finish().await.unwrap(); // close the client control stream immediately
+        control_stream.finish().unwrap(); // close the client control stream immediately
 
-        let (mut driver, _send) = client::new(h3_quinn::Connection::new(connection))
-            .await
-            .unwrap();
+        // create the Connection manually, so it does not open a second Control stream
 
-        future::poll_fn(|cx| driver.poll_close(cx)).await
+        let connection_error = loop {
+            let accepted = connection.accept_bi().await;
+            match accepted {
+                // do nothing with the stream
+                Ok(_) => continue,
+                Err(err) => break err,
+            }
+        };
+
+        let err_code = match connection_error {
+            quinn::ConnectionError::ApplicationClosed(quinn::ApplicationClose {
+                error_code,
+                ..
+            }) => error_code.into_inner(),
+            e => panic!("unexpected error: {:?}", e),
+        };
+        assert_eq!(err_code, Code::H3_CLOSED_CRITICAL_STREAM.value());
     };
 
     let server_fut = async {
@@ -354,7 +421,7 @@ async fn control_close_send_error() {
             if *reason == *"control stream closed");
     };
 
-    tokio::select! { _ = server_fut => (), _ = client_fut => panic!("client resolved first") };
+    tokio::join!(server_fut, client_fut);
 }
 
 #[tokio::test]
@@ -472,7 +539,7 @@ async fn goaway_from_server_not_request_id() {
         let mut buf = BytesMut::new();
         StreamType::CONTROL.encode(&mut buf);
         control_stream.write_all(&buf[..]).await.unwrap();
-        control_stream.finish().await.unwrap(); // close the client control stream immediately
+        control_stream.finish().unwrap(); // close the client control stream immediately
 
         let (mut driver, _send) = client::new(h3_quinn::Connection::new(connection))
             .await
